require "virtus"
require "active_support/inflector"
require "active_support/core_ext/object/try"

module DataBroker
  def self.value_object
    Module.new do
      def self.included(descendant)
        descendant.send(:include, ::DataBroker::ValueObject)
      end
    end
  end

  module ValueObject
    def self.included(descendant)
      descendant.send(:include, Virtus.value_object)
      descendant.send(:include, ActiveSupport::Inflector)
      descendant.send(:include, InstanceMethods)
      descendant.extend(ClassMethods)
    end

    module InstanceMethods
      def method_missing(method, *args, &block)
        if method == :type
          self.class.to_s
        else
          super
        end
      end

      def respond_to_missing?(method, include_private = false)
        method == :type || super
      end
    end

    module ClassMethods
      include ActiveSupport::Inflector

      def belongs_to(method)
        mapper = "#{camelize(method.to_s)}Mapper"
        via = foreign_key(camelize(method.to_s))

        define_method(method) do
          ivar = "@#{method}"
          instance_variable_get(ivar) || instance_variable_set(
            ivar, constantize(mapper).find(self.send(via))
          )
        end
      end

      def has_many(method)
        mapper = "#{classify(method.to_s)}Mapper"
        via = foreign_key(self.to_s)
        ids_method = pluralize(foreign_key(classify(method.to_s)))

        define_method(method) do
          ivar = "@#{method}"
          instance_variable_get(ivar) || -> {
            instance_variable_set("@#{ids_method}", nil)
            instance_variable_set(
              ivar, constantize(mapper).where(via.to_sym => self.id)
            )
          }.call
        end

        define_method(ids_method) do
          ivar = "@#{ids_method}"
          instance_variable_get(ivar) || instance_variable_set(
            ivar, send(method).map(&:id)
          )
        end
      end

      def delegate(method, to: nil, return_on_nil: nil)
        delegated = "#{to}_#{method}"
        define_method(delegated) do
          ivar = "@#{delegated}"
          instance_variable_get(ivar) ||
            instance_variable_set(ivar, send(to).try(method)) ||
            return_on_nil
        end
      end
    end
  end
end
